Package: UCSC.utils
Title: Low-level utilities to retrieve data from the UCSC Genome
        Browser
Description: A set of low-level utilities to retrieve data from the UCSC
	Genome Browser. Most functions in the package access the data via
	the UCSC REST API but some of them query the UCSC MySQL server
	directly. Note that the primary purpose of the package is to support
	higher-level functionalities implemented in downstream packages
	like GenomeInfoDb or txdbmaker.
biocViews: Infrastructure, GenomeAssembly, Annotation,
        GenomeAnnotation, DataImport
URL: https://bioconductor.org/packages/UCSC.utils
BugReports: https://github.com/Bioconductor/UCSC.utils/issues
Version: 1.9.0
License: Artistic-2.0
Encoding: UTF-8
Authors@R: person("Hervé", "Pagès", role=c("aut", "cre"),
	          email="hpages.on.github@gmail.com")
Imports: methods, stats, httr, jsonlite, S4Vectors (>= 0.47.6)
Suggests: DBI, RMariaDB, GenomeInfoDb, testthat, knitr, rmarkdown,
        BiocStyle
VignetteBuilder: knitr
Collate: 00utils.R UCSC.api.url.R REST_API.R list_UCSC_genomes.R
        get_UCSC_chrom_sizes.R list_UCSC_tracks.R
        fetch_UCSC_track_data.R UCSC_dbselect.R zzz.R
git_url: https://git.bioconductor.org/packages/UCSC.utils
git_branch: devel
git_last_commit: 5a1b81d
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.24
Date/Publication: 2026-04-28
NeedsCompilation: no
Packaged: 2026-04-29 01:42:17 UTC; biocbuild
Author: Hervé Pagès [aut, cre]
Maintainer: Hervé Pagès <hpages.on.github@gmail.com>

import(methods)
importFrom(stats, setNames)
importFrom(httr, GET, content, user_agent, stop_for_status)
importFrom(jsonlite, fromJSON)

importFrom(S4Vectors, wmsg, isTRUEorFALSE, isSingleNumber,
                      isSingleString, isSingleStringOrNA,
                      toListOfIntegerVectors)

export(
    UCSC.api.url,
    list_UCSC_genomes,
    get_organism_for_UCSC_genome,
    get_UCSC_chrom_sizes,
    list_UCSC_tracks,
    fetch_UCSC_track_data,
    UCSC_dbselect
)

### =========================================================================
### fetch_UCSC_track_data()
### -------------------------------------------------------------------------
###


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### .extract_table_data_from_parsed_json()
###
### WARNING: For JSON parsed with rjson::fromJSON()!
###
### Lots of massaging and sanity checks to deal with the messiness of UCSC's
### JSON.
###

.extract_table_data_from_parsed_json <- function(parsed_json, primary_table)
{
    stopifnot(is.list(parsed_json), !is.null(names(parsed_json)))
    table_data <- parsed_json[[primary_table]]
    ## The table data is either put all together in a single table or split
    ## into one table per chromosome. If the former then 'table_data' is an
    ## unnamed list with one list element per row in the table. If the latter
    ## then 'table_data' is a named list with one list element per chromosome,
    ## where each list element represents a table.
    stopifnot(is.list(table_data))
    if (is.null(names(table_data))) {
        ## One single table.
        ans <- make_data_frame_from_list_of_rows(table_data)
    } else {
        ## 'table_data' is a named list with the chromosome names on it. Each
        ## list element in 'table_data' is itself a list that represents a
        ## table.
        idx <- which(lengths(table_data) != 0L)
        if (length(idx) == 0L) {
            ## We don't know what the columns would have been if the track
            ## had some data so we return a 0-col data frame.
            ## Happens for example with fetch_UCSC_track_data("eboVir3", "gap").
            warning(wmsg("track is empty ==> returning a 0x0 data frame"))
            return(data.frame())
        }
        dfs <- lapply(table_data[idx], make_data_frame_from_list_of_rows)
        ans <- do.call(rbind, unname(dfs))
    }
    stopifnot(nrow(ans) == parsed_json[["itemsReturned"]])
    ans
}


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### .extract_table_data_from_parsed_json2()
###
### WARNING: For JSON parsed with jsonlite::fromJSON()!
###
### A slightly simpler alternative to .extract_table_data_from_parsed_json()
### that takes advantage of the built-in data massaging capabilities of
### jsonlite::fromJSON().
### NOT a drop-in replacement for .extract_table_data_from_parsed_json() as
### it first requires switching from rjson::fromJSON to jsonlite::fromJSON
### in internal helper .parse_json() defined in R/REST_API.R.
### As Marcel pointed out, unlike the former the latter recognizes JSON
### lists that represent tabular data and automatically turns them into data
### frames. See https://github.com/Bioconductor/Contributions/issues/3343
###
### Note that the code below still needs to perform a little bit of data
### massaging and sanity checks to deal with the messiness of UCSC's JSON!
###

### The table data is either put all together in a single data frame or split
### into one data frame per chromosome.
### Make sure to always return a single data frame.
.table_data_as_data_frame <- function(table_data)
{
    if (is.data.frame(table_data))
        return(table_data)
    ## 'table_data' should be either an empty list or a named list of
    ## data frames (one per chromosome).
    stopifnot(is.list(table_data))
    if (length(table_data) == 0L) {
        warning(wmsg("track is empty ==> returning a 0x0 data frame"))
        return(data.frame())
    }
    ## From now on, 'table_data' is expected to be a named list of data frames,
    ## one per chromosome.
    stopifnot(!is.null(names(table_data)))
    if (length(table_data) != 1L)
        return(do.call(rbind, unname(table_data)))
    ans <- table_data[[1L]]
    if (length(ans) == 0L) {
        warning(wmsg("track is empty ==> returning a 0x0 data frame"))
        return(data.frame())
    }
    ans
}

.extract_table_data_from_parsed_json2 <- function(parsed_json, primary_table)
{
    stopifnot(is.list(parsed_json), !is.null(names(parsed_json)))
    table_data <- parsed_json[[primary_table]]
    ans <- .table_data_as_data_frame(table_data)
    stopifnot(nrow(ans) == parsed_json[["itemsReturned"]])
    ans
}


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### fetch_UCSC_track_data()
###

### No caching!
fetch_UCSC_track_data <- function(genome, primary_table, api.url=UCSC.api.url())
{
    check_genome(genome)
    if (!(isSingleString(primary_table) && nzchar(primary_table)))
        stop(wmsg("'primary_table' must be a single (non-empty) string"))
    parsed_json <- API_get_track_data(genome, primary_table, api.url=api.url)
    .extract_table_data_from_parsed_json2(parsed_json, primary_table)
}

test_that("fetch_UCSC_track_data()", {
    result <- fetch_UCSC_track_data("gorGor6", "ncbiRefSeqCurated")
    expect_true(is.data.frame(result))
    EXPECTED_COLNAMES <- c("bin", "name", "chrom", "strand",
                           "txStart", "txEnd", "cdsStart", "cdsEnd",
                           "exonCount", "exonStarts", "exonEnds", "score",
                           "name2", "cdsStartStat", "cdsEndStat", "exonFrames")
    expect_identical(colnames(result), EXPECTED_COLNAMES)
    expect_true(nrow(result) >= 400L)

    result <- fetch_UCSC_track_data("gorGor6", "gap")
    expect_true(is.data.frame(result))
    EXPECTED_COLNAMES <- c("bin", "chrom", "chromStart", "chromEnd", "ix" ,                                "n", "size", "type", "bridge")
    expect_identical(colnames(result), EXPECTED_COLNAMES)
    expect_true(nrow(result) >= 800L)

    result <- fetch_UCSC_track_data("eboVir3", "iedbBcell")
    expect_true(is.data.frame(result))
    expect_true(ncol(result) == 76L)
    expect_true(nrow(result) >= 50L)
})

### =========================================================================
### UCSC.api.url()
### -------------------------------------------------------------------------
###

### See https://genome.ucsc.edu/goldenPath/help/api.html#Mirrors
.UCSC_API_URLS <- c(
  primary="https://api.genome.ucsc.edu",  # primary URL (West Coast)
  europe="https://genome-euro.ucsc.edu/cgi-bin/hubApi",
  asia="https://genome-asia.ucsc.edu/cgi-bin/hubApi"
)

.pattern2url <- function(pattern)
{
    if (!nzchar(pattern))
        return(.UCSC_API_URLS["primary"])
    aliases <- names(.UCSC_API_URLS)
    idx <- grep(pattern, aliases, ignore.case=TRUE)
    if (length(idx) == 0L) {
        in1string <- paste0("\"", c("", aliases), "\"")
        in1string[[length(in1string)]] <-
            paste0("or ", in1string[[length(in1string)]])
        in1string <- paste(in1string, collapse=", ")
        stop(wmsg("'api.url' must start with \"https://\" (or \"http://\") ",
                  "or match alias ", in1string))
    }
    if (length(idx) > 1L) {
        in1string <- paste0("\"", aliases[idx], "\"", collapse=" and ")
        stop(wmsg("\"", tolower(pattern), "\" matches more than ",
                  "one alias (", in1string, ")"))
    }
    .UCSC_API_URLS[idx]
}

normarg_api.url <- function(api.url)
{
    if (!isSingleString(api.url))
        stop(wmsg("'api.url' must be a single string"))
    if (grepl("^https?://", api.url, ignore.case=TRUE))
        return(api.url)
    .pattern2url(api.url)
}

UCSC.api.url <- function(api.url=NULL)
{
    ans <- getOption("UCSC.api.url")
    if (is.null(api.url))
        return(ans)
    api.url <- normarg_api.url(api.url)
    options(UCSC.api.url=api.url)
    invisible(ans)  # return previous URL invisibly
}

### =========================================================================
### list_UCSC_genomes() and get_organism_for_UCSC_genome()
### -------------------------------------------------------------------------
###


.list_UCSC_genomes <- function(api.url=UCSC.api.url(), recache=FALSE)
{
    if (!isTRUEorFALSE(recache))
        stop(wmsg("'recache' must be TRUE or FALSE"))
    key <- "GENOMES"
    ans <- cached_rest_api_results[[key]]
    if (is.null(ans) || recache) {
        ans <- API_list_genomes(api.url=api.url)
        cached_rest_api_results[[key]] <- ans
    }
    ans
}

### NOT exported but used in GenomeInfoDb::registered_UCSC_genomes().
### Orders first by 'organism' then by 'genome'.
order_organism_genome_pairs <- function(organism, genome)
{
    regexpr <- "^(.*[^0-9])([0-9]*)$"
    genome_basename <- sub(regexpr, "\\1", genome)
    genome_version <- as.integer(sub(regexpr, "\\2", genome))
    genome_version[is.na(genome_version)] <- 0L
    order(organism, genome_basename, genome_version)
}

### Returns a data.frame with 1 row per genome and 5 columns:
### organism, genome, common_name, tax_id, description.
### A few things align with GenomeInfoDb::registered_UCSC_genomes():
###   - colnames "organism" and "genome" on the returned data frame;
###   - column "organism" returned as a factor;
###   - name, default value, and behavior of 'organism' argument, but with
###     the difference that we also search matches in the "common_name"
###     column (in addition to matches in the "organism" column);
###   - order of rows in the returned data frame.
list_UCSC_genomes <- function(organism=NA, api.url=UCSC.api.url(),
                                           recache=FALSE)
{
    if (!isSingleStringOrNA(organism))
        stop(wmsg("'organism' must be a single string or NA"))
    genomes <- .list_UCSC_genomes(api.url=api.url, recache=recache)

    ans_organism <- factor(vapply(genomes,
        function(genome) {
            stopifnot(is.list(genome))  # sanity check
            genome[["scientificName"]]
        },
        character(1), USE.NAMES=FALSE
    ))
    ans_common_name <- factor(vapply(genomes,
        function(genome) genome[["organism"]],
        character(1), USE.NAMES=FALSE
    ))
    if (!is.na(organism)) {
        keep_idx <- which(grepl(organism, ans_organism, ignore.case=TRUE) |
                          grepl(organism, ans_common_name, ignore.case=TRUE))
        genomes <- genomes[keep_idx]
        ans_organism <- ans_organism[keep_idx]
        ans_common_name <- ans_common_name[keep_idx]
    }

    ans_genome <- names(genomes)
    ans_tax_id <- vapply(genomes,
        function(genome) as.integer(genome[["taxId"]]),
        integer(1), USE.NAMES=FALSE
    )
    ans_description <- vapply(genomes,
        function(genome) genome[["description"]],
        character(1), USE.NAMES=FALSE
    )

    ans <- data.frame(
        organism=ans_organism,
        genome=ans_genome,
        common_name=ans_common_name,
        tax_id=ans_tax_id,
        description=ans_description
    )
    oo <- order_organism_genome_pairs(ans[ , "organism"], ans[ , "genome"])
    S4Vectors:::extract_data_frame_rows(ans, oo)
}


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### get_organism_for_UCSC_genome()
###

### A convenience helper based on list_UCSC_genomes(). Vectorized.
get_organism_for_UCSC_genome <- function(genome, api.url=UCSC.api.url(),
                                                 recache=FALSE)
{
    if (!is.character(genome))
        stop(wmsg("'genome' must be a character vector"))
    if (anyNA(genome) || !all(nzchar(genome))) {
        if (length(genome) == 1L)
            msg <- "be NA or the empty string"
        else
            msg <- "contain NAs or empty strings"
        stop(wmsg("'genome' cannot ", msg))
    }
    df <- list_UCSC_genomes(api.url=api.url, recache=recache)
    idx <- match(genome, df[ , "genome"])
    if (anyNA(idx)) {
        bad_genomes <- genome[is.na(idx)]
        if (length(bad_genomes) == 1L)
            stop(wmsg(bad_genomes, ": unknown UCSC genome"))
        bad_genomes <- paste(bad_genomes, collapse=",")
        stop(wmsg(bad_genomes, ": unknown UCSC genomes"))
    }
    setNames(as.character(df[idx, "organism"]), genome)
}


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Internal helper check_genome()
###

### Not exported.
check_genome <- function(genome)
{
    if (!(isSingleString(genome) && nzchar(genome)))
        stop(wmsg("'genome' must be a single (non-empty) string"))
    if (!(genome %in% list_UCSC_genomes()[ , "genome"]))
        stop(wmsg(genome, ": unknown UCSC genome "))
}

.check_list_UCSC_genomes_result <- function(df) {
    expect_true(is.data.frame(df))
    EXPECTED_COLNAMES <- c("organism", "genome", "common_name",
                           "tax_id", "description")
    expect_identical(colnames(df), EXPECTED_COLNAMES)
    expect_true(is.factor(df$organism))
    expect_true(is.character(df$genome))
    expect_true(is.factor(df$common_name))
    expect_true(is.integer(df$tax_id))
    expect_true(is.character(df$description))
}

test_that("list_UCSC_genomes()", {
    result <- list_UCSC_genomes("cat")
    .check_list_UCSC_genomes_result(result)
    expect_true(nrow(result) >= 2L)

    result <- list_UCSC_genomes("___XXX___XXX___")
    .check_list_UCSC_genomes_result(result)

    expect_error(list_UCSC_genomes(organism=NULL))
    expect_error(list_UCSC_genomes(organism=22))
})

test_that("get_organism_for_UCSC_genome()", {
    genomes <- c("hs1", "hg38", "dm6", "ce11", "mm39",
                 "dm6", "xenTro10", "mpxvRivers")
    expected <- c("Homo sapiens", "Homo sapiens", "Drosophila melanogaster",
                  "Caenorhabditis elegans", "Mus musculus",
                  "Drosophila melanogaster", "Xenopus tropicalis",
                  "Monkeypox virus")
    names(expected) <- genomes
    result <- get_organism_for_UCSC_genome(genomes)
    expect_identical(result, expected)

    result <- get_organism_for_UCSC_genome(character(0))
    expect_identical(result, setNames(character(0), character(0)))

    expect_error(get_organism_for_UCSC_genome(NULL))
    expect_error(get_organism_for_UCSC_genome(22))
    expect_error(get_organism_for_UCSC_genome(""))
    expect_error(get_organism_for_UCSC_genome(NA))
    expect_error(get_organism_for_UCSC_genome("_not_a_genome_"))
})

### =========================================================================
### list_UCSC_tracks()
### -------------------------------------------------------------------------
###


.get_UCSC_genome_tracks <- function(genome, api.url=UCSC.api.url(),
                                            recache=FALSE)
{
    check_genome(genome)
    if (!isTRUEorFALSE(recache))
        stop(wmsg("'recache' must be TRUE or FALSE"))
    key <- paste0(genome, "_TRACKS")
    ans <- cached_rest_api_results[[key]]
    if (is.null(ans) || recache) {
        ans <- API_list_tracks(genome, api.url=api.url)
        cached_rest_api_results[[key]] <- ans
    }
    ans
}

### Typical usage:
###     list_UCSC_tracks("ce2")
###     list_UCSC_tracks("mm9", group="genes")
###     list_UCSC_tracks("hg38", group=NA)
### Returns a data.frame with 1 row per track and 5 columns:
### track, primary_table, type, group, composite_track.
### Note that columns "group" and "composite_track" can contain NAs.
### Passing 'group=NA' is accepted and keeps only rows for tracks that
### don't belong to any group. This is why default value for the 'group'
### argument is NULL and not NA like for the 'organism' argument in
### list_UCSC_genomes() above.
list_UCSC_tracks <- function(genome, group=NULL,
                             api.url=UCSC.api.url(), recache=FALSE)
{
    if (!(is.null(group) || isSingleStringOrNA(group)))
        stop(wmsg("'group' must be a single string, or NA, or NULL"))
    genome_tracks <- .get_UCSC_genome_tracks(genome, api.url=api.url,
                                                     recache=recache)

    track_groups <- factor(vapply(genome_tracks,
        function(track) {
            stopifnot(is.list(track))  # sanity check
            idx <- match("group", names(track))
            if (is.na(idx)) NA_character_ else track[[idx]]
        },
        character(1), USE.NAMES=FALSE
    ), exclude=character(0))
    if (!is.null(group)) {
        keep_idx <- which(track_groups %in% group)
        genome_tracks <- genome_tracks[keep_idx]
        track_groups <- track_groups[keep_idx]
    }

    track_shortlabels <- vapply(genome_tracks,
        function(track) track[["shortLabel"]],
        character(1), USE.NAMES=FALSE
    )
    track_types <- vapply(genome_tracks,
        function(track) track[["type"]],
        character(1), USE.NAMES=FALSE
    )

    ## Extract tracks nested in composite tracks.
    is_composite <- vapply(genome_tracks,
        function(track) identical(track[["compositeTrack"]], "on"),
        logical(1), USE.NAMES=FALSE
    )
    nested_tracks <- lapply(genome_tracks[is_composite],
        function(track) {
            track[vapply(track, is.list, logical(1), USE.NAMES=FALSE)]
        }
    )
    nested_track_shortlabels <- lapply(nested_tracks,
        function(tracks) vapply(tracks,
                                function(track) track[["shortLabel"]],
                                character(1), USE.NAMES=FALSE)
    )
    nested_primary_tables <- lapply(nested_tracks, names)
    nested_track_types <- lapply(nested_tracks,
        function(tracks) vapply(tracks,
                                function(track) track[["type"]],
                                character(1), USE.NAMES=FALSE)
    )
    nested_tracks_count <- lengths(nested_tracks)

    ## Sanity checks.
    stopifnot(
        identical(lengths(nested_track_shortlabels), nested_tracks_count),
        identical(lengths(nested_primary_tables), nested_tracks_count),
        identical(lengths(nested_track_types), nested_tracks_count)
    )

    ## Prepare columns of final data frame.
    times <- rep.int(1L, length(genome_tracks))
    times[is_composite] <-  nested_tracks_count
    ans_is_composite <- rep.int(is_composite, times)
    ans_track <- ans_composite_track <- rep.int(track_shortlabels, times)
    ans_track[ans_is_composite] <-
        unlist(nested_track_shortlabels, use.names=FALSE)
    ans_primary_table <- rep.int(names(genome_tracks), times)
    ans_primary_table[ans_is_composite] <-
        unlist(nested_primary_tables, use.names=FALSE)
    stopifnot(anyDuplicated(ans_primary_table) == 0L)  # sanity check
    ans_type <- rep.int(track_types, times)
    ans_type[ans_is_composite] <-
        unlist(nested_track_types, use.names=FALSE)
    ans_group <- rep.int(track_groups, times)
    ans_composite_track[!ans_is_composite] <- NA_character_

    data.frame(
        track=ans_track,
        primary_table=ans_primary_table,
        type=ans_type,
        group=ans_group,
        composite_track=ans_composite_track
    )
}

.check_list_UCSC_tracks_result <- function(df) {
    expect_true(is.data.frame(df))
    EXPECTED_COLNAMES <- c("track", "primary_table", "type",
                           "group", "composite_track")
    expect_identical(colnames(df), EXPECTED_COLNAMES)
    expect_true(is.character(df$track))
    expect_true(is.character(df$primary_table))
    expect_true(is.character(df$type))
    expect_true(is.factor(df$group))
    expect_true(is.character(df$composite_track))
}

test_that("list_UCSC_tracks()", {
    result <- list_UCSC_tracks("ce11")
    .check_list_UCSC_tracks_result(result)

    result <- list_UCSC_tracks("ce11", group="genes")
    .check_list_UCSC_tracks_result(result)
    expect_true(nrow(result) >= 2L)
    expect_true(all(result$group == "genes"))

    result <- list_UCSC_tracks("ce11", group="_not_a_group_")
    .check_list_UCSC_tracks_result(result)

    result <- list_UCSC_tracks("hg38", group=NA)
    .check_list_UCSC_tracks_result(result)
    expect_true(nrow(result) >= 2L)
    expect_true(is.na(tail(levels(result$group), n=1L)))
    expect_true(all(is.na(as.character(result$group))))

    expect_error(list_UCSC_tracks(NULL))
    expect_error(list_UCSC_tracks(22))
    expect_error(list_UCSC_tracks(""))
    expect_error(list_UCSC_tracks(NA))
    expect_error(list_UCSC_tracks("_not_a_genome_"))
    expect_error(list_UCSC_tracks("ce11", group=22))
})

### =========================================================================
### Thin R wrappers to UCSC REST API endpoints
### -------------------------------------------------------------------------
###
### Nothing in this file is exported.
###


.parse_json <- function(response)
{
    ## If the JSON in the content of the response is encoded with
    ## Windows-1252, then content(response, as="text", encoding="UTF-8")
    ## will silently return an NA (see ?httr::content). This happens for
    ## example with the following query:
    ##   query <- list(genome="eboVir3", track="iedbBcell")
    ##   .query_API("getData/track", query=query)
    ## This query returns a response with bytes 233 (\xe9) and 246 (\xf6)
    ## in response$content. These bytes cause the call to content() below
    ## to silently return an NA.
    json_string <- content(response, as="text", encoding="UTF-8")
    stopifnot(is.character(json_string), length(json_string) == 1L)
    if (is.na(json_string)) {
        json_string <- content(response, as="text", encoding="Windows-1252")
        stopifnot(isSingleString(json_string))
    }
    parsed_json <- jsonlite::fromJSON(json_string)
    ## Sanity checks.
    stopifnot(is.list(parsed_json), !is.null(names(parsed_json)))
    parsed_json
}

.check_response_status <- function(response, fallback_errmsg=NULL)
{
    status_code <- response[["status_code"]]
    ## HTTP errors 400 and 415 still return parsable JSON with valuable
    ## details about the error.
    ## Error 400 happens for example with the following query:
    ##   query <- list(genome="mm9", track="chainNetBosTau6Viewnet")
    ##   .query_API("getData/track", query=query)
    ## Error 415 happens for example with the following query:
    ##   query <- list(genome="mm9", track="bamMmsNumtSSorted")
    ##   .query_API("getData/track", query=query)
    if (status_code %in% c(400L, 415L)) {
        parsed_json <- .parse_json(response)
        stop(wmsg("[HTTP ", status_code, "] ",
                  parsed_json$statusMessage, ": ", parsed_json$error))
    }
    if (is.null(fallback_errmsg)) {
        stop_for_status(response)
        return(invisible(NULL))
    }
    if (status_code >= 300L)
        stop(wmsg("[HTTP ", status_code, "] ", fallback_errmsg))
}

### Returns the parsed JSON content of the response.
.query_API <- function(endpoint, query=list(), api.url=UCSC.api.url(),
                       fallback_errmsg=NULL)
{
    stopifnot(isSingleString(endpoint), nzchar(endpoint),
              is.list(query),
              isSingleString(api.url), nzchar(api.url))
    if (length(query) != 0L)
        stopifnot(!is.null(names(query)))
    url <- paste0(api.url, "/", endpoint)
    response <- GET(url, user_agent("Bioconductor UCSC.utils"), query=query)
    .check_response_status(response, fallback_errmsg)
    .parse_json(response)
}


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### We only support the following endpoints at the moment:
###   - /list/ucscGenomes
###   - /list/chromosomes
###   - /list/tracks
###   - /getData/track
### See https://genome.ucsc.edu/goldenPath/help/api.html#Endpoint for the
### full list of endpoints.
###
### All the functions below return parsed JSON.
###

### Endpoint /list/ucscGenomes
API_list_genomes <- function(api.url=UCSC.api.url())
{
    api.url <- normarg_api.url(api.url)

    endpoint <- "list/ucscGenomes"
    fallback_errmsg <- c("failed to get list of UCSC genomes from ", api.url)
    parsed_json <- .query_API(endpoint, api.url=api.url,
                              fallback_errmsg=fallback_errmsg)
    ans <- parsed_json[["ucscGenomes"]]
    ## Sanity check.
    stopifnot(is.list(ans))
    ans
}

### Endpoint /list/chromosomes
API_list_chromosomes <- function(genome, api.url=UCSC.api.url())
{
    stopifnot(isSingleString(genome), nzchar(genome))
    api.url <- normarg_api.url(api.url)

    endpoint <- "list/chromosomes"
    query <- list(genome=genome)
    fallback_errmsg <- c(genome, ": unknown UCSC genome ",
                         "(or ", api.url, " is down?)")
    parsed_json <- .query_API(endpoint, query=query, api.url=api.url,
                              fallback_errmsg=fallback_errmsg)
    ## Sanity check.
    stopifnot(identical(parsed_json[["genome"]], genome))
    parsed_json
}

### Endpoint /list/tracks
API_list_tracks <- function(genome, api.url=UCSC.api.url())
{
    stopifnot(isSingleString(genome), nzchar(genome))
    api.url <- normarg_api.url(api.url)

    endpoint <- "list/tracks"
    query <- list(genome=genome)
    fallback_errmsg <- c(genome, ": unknown UCSC genome ",
                         "(or ", api.url, " is down?)")
    parsed_json <- .query_API(endpoint, query=query, api.url=api.url,
                              fallback_errmsg=fallback_errmsg)
    ans <- parsed_json[[genome]]
    ## Sanity check.
    stopifnot(is.list(ans))
    ans
}

### Endpoint /getData/track
### Note that the endpoint expects the supplied 'track' argument to be the
### name of the track's primary table rather than the track's name.
### E.g. "catLiftOffGenesV1" rather than "CAT/Liftoff Genes".
API_get_track_data <- function(genome, primary_table, api.url=UCSC.api.url())
{
    stopifnot(isSingleString(genome), nzchar(genome),
              isSingleString(primary_table), nzchar(primary_table))
    api.url <- normarg_api.url(api.url)

    endpoint <- "getData/track"
    query <- list(genome=genome, track=primary_table)
    fallback_errmsg <- c(genome, "/", primary_table, ": ",
                         "unknown UCSC genome/primary_table ",
                         "(or ", api.url, " is down?)")
    parsed_json <- .query_API(endpoint, query=query, api.url=api.url,
                              fallback_errmsg=fallback_errmsg)
    ## Sanity checks.
    stopifnot(identical(parsed_json[["genome"]], genome))
    if (!is.null(parsed_json[["track"]]))
        stopifnot(identical(parsed_json[["track"]], primary_table))
    parsed_json
}

cached_rest_api_results <- new.env(parent=emptyenv())


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### make_data_frame_from_list_of_rows()
###

.lossless_num_to_int <- function(x)
{
    stopifnot(is.numeric(x))
    y <- suppressWarnings(as.integer(x))
    if (identical(as.numeric(y), x)) y else x
}

### Looks like tabular data in JSON is usually row-oriented (i.e. one list
### element per row, each row itself being represented by a named list of
### length-1 atomic vectors), instead of column-oriented. At least that's
### how UCSC's /getData/track endpoint returns their track data.
### Soooo inefficient!
make_data_frame_from_list_of_rows <- function(list_of_rows)
{
    stopifnot(is.list(list_of_rows), is.null(names(list_of_rows)))
    if (length(list_of_rows) == 0L) {
        ## Happens for example with
        ## fetch_UCSC_track_data("eboVir3", "unipAliSwissprot").
        warning(wmsg("track is empty ==> returning a 0x0 data frame"))
        return(data.frame())
    }

    ## Turn list of rows into list of columns (transposition).

    ## 1st implementation.
    #ans_colnames <- names(list_of_rows[[1L]])
    #list_of_cols <- lapply(setNames(ans_colnames, ans_colnames),
    #    function(colname) {
    #        col <- sapply(list_of_rows, function(row) row[[colname]],
    #                      USE.NAMES=FALSE)
    #        if (is.numeric(col))
    #            col <- .lossless_num_to_int(col)
    #        col
    #    }
    #)

    ## 2nd implementation: About 3x faster than the above!
    ## Assumes that all list elements in 'list_of_rows' are ordered the
    ## same i.e. have the same names in the same order.
    #ans_ncol <- length(list_of_rows[[1L]])
    #stopifnot(all(lengths(list_of_rows) == ans_ncol))
    ### Turn 'list_of_rows' into a matrix of type list with
    ### 'length(list_of_rows)' rows and 'ans_ncol' cols.
    #m <- do.call(rbind, list_of_rows)
    #ans_colnames <- colnames(m)
    #list_of_cols <- lapply(setNames(seq_len(ans_ncol), ans_colnames),
    #    function(j) {
    #        col <- unlist(m[ , j], recursive=FALSE, use.names=FALSE)
    #        if (is.numeric(col))
    #            col <- .lossless_num_to_int(col)
    #        col
    #    }
    #)

    ## 3rd implementation: Even slightly faster than 2nd implementation.
    ## Also assumes that all list elements in 'list_of_rows' are ordered
    ## the same i.e. have the same names in the same order.
    ans_ncol <- length(list_of_rows[[1L]])
    ans_colnames <- names(list_of_rows[[1L]])
    stopifnot(all(lengths(list_of_rows) == ans_ncol))
    ## Turn 'list_of_rows' into a matrix of type list with
    ## 'ans_ncol' rows and 'length(list_of_rows)' cols.
    unlisted <- unlist(list_of_rows, recursive=FALSE, use.names=FALSE)
    m <- matrix(unlisted, nrow=ans_ncol)
    list_of_cols <- lapply(setNames(seq_len(ans_ncol), ans_colnames),
        function(i) {
            col <- unlist(m[i, ], recursive=FALSE, use.names=FALSE)
            if (is.numeric(col))
                col <- .lossless_num_to_int(col)
            col
        }
    )

    as.data.frame(list_of_cols, check.names=FALSE)
}

### =========================================================================
### get_UCSC_chrom_sizes()
### -------------------------------------------------------------------------
### 


### Returns the chromosome sizes in an named numeric vector.
.extract_chrom_sizes_from_parsed_json <- function(parsed_json)
{
    stopifnot(is.list(parsed_json), !is.null(names(parsed_json)))
    chrom_count <- parsed_json[["chromCount"]]
    stopifnot(isSingleNumber(chrom_count))
    chrom_count <- as.integer(chrom_count)
    chrom_sizes <- parsed_json[["chromosomes"]]
    stopifnot(is.list(chrom_sizes),
              !is.null(names(chrom_sizes)),
              length(chrom_sizes) == chrom_count)
    chrom_sizes <- unlist(chrom_sizes, recursive=FALSE)
    stopifnot(is.numeric(chrom_sizes),
              length(chrom_sizes) == chrom_count)
    chrom_sizes
}

get_UCSC_chrom_sizes <- function(genome, api.url=UCSC.api.url(), recache=FALSE)
{
    check_genome(genome)
    if (!isTRUEorFALSE(recache))
        stop(wmsg("'recache' must be TRUE or FALSE"))
    key <- paste0(genome, "_CHROMOSOMES")
    ans <- cached_rest_api_results[[key]]
    if (is.null(ans) || recache) {
        parsed_json <- API_list_chromosomes(genome, api.url=api.url)
        ans <- .extract_chrom_sizes_from_parsed_json(parsed_json)
        cached_rest_api_results[[key]] <- ans
    }
    ans
}

test_that("get_UCSC_chrom_sizes()", {
    result <- get_UCSC_chrom_sizes("ce11")
    expect_true(is.numeric(result))
    expect_false(is.null(names(result)))

    expect_error(get_UCSC_chrom_sizes(NULL))
    expect_error(get_UCSC_chrom_sizes(22))
    expect_error(get_UCSC_chrom_sizes(""))
    expect_error(get_UCSC_chrom_sizes(NA))
    expect_error(get_UCSC_chrom_sizes("_not_a_genome_"))

})

### =========================================================================
### UCSC_dbselect()
### -------------------------------------------------------------------------
###


.make_SQL_SELECT <- function(from, columns=NULL, where=NULL, MoreSQL=NULL)
{
    if (!(isSingleString(from) && nzchar(from)))
        stop(wmsg("'from' must be a single (non-empty) string"))

    if (is.null(columns)) {
        columns <- "*"
    } else {
        if (!is.character(columns))
            stop(wmsg("'columns' must be NULL or a character vector"))
        if (length(columns) == 0L)
            stop(wmsg("'columns' cannot have length zero"))
        if (anyNA(columns) || !all(nzchar(columns)))
            stop(wmsg("'columns' cannot contain NAs or empty strings"))
        columns <- paste0(columns, collapse=",")
    }

    SQL <- sprintf("SELECT %s FROM %s", columns, from)

    if (!is.null(where)) {
        if (!(isSingleString(where) && nzchar(where)))
            stop(wmsg("'where' must be NULL or a single (non-empty) string"))
        SQL <- paste(SQL, "WHERE", where)
    }

    if (!is.null(MoreSQL)) {
        if (!is.character(MoreSQL))
            stop(wmsg("'MoreSQL' must be NULL or a character vector"))
        if (length(MoreSQL) != 0L)
            SQL <- paste(SQL, paste(MoreSQL, collapse=" "))
    }
    SQL
}

.fix_RMariaDB_blobs <- function(df, try.blob.as.list=FALSE)
{
    stopifnot(is.data.frame(df), isTRUEorFALSE(try.blob.as.list))
    idx <- which(vapply(df, inherits, logical(1), "blob"))
    df[idx] <- lapply(df[idx],
        function(col) {
            if (try.blob.as.list) {
                col2 <- try(toListOfIntegerVectors(col), silent=TRUE)
                if (!inherits(col2, "try-error"))
                    return(col2)
            }
            vapply(col, rawToChar, character(1), USE.NAMES=FALSE)
        }
    )
    df
}

.unmangle_RMariaDB_colnames <- function(colnames)
{
    stopifnot(is.character(colnames))
    ncol <- length(colnames)
    if (ncol == 0L)
        return(colnames)

    ## Identify columns with the dumb suffixes.
    dumb_suffixes <- paste0("..", seq_len(ncol))
    nc1 <- nchar(colnames)
    nc2 <- nchar(dumb_suffixes)
    suffixes <- substr(colnames, start=nc1-nc2+1L, stop=nc1)
    idx <- which(suffixes == dumb_suffixes)

    ## Remove the dumb suffixes.
    colnames[idx] <- substr(colnames[idx], start=1L, stop=nc1[idx]-nc2[idx])

    colnames
}

### See https://genome.ucsc.edu/goldenpath/help/mysql.html for how to connect
### to a MariaDB server at UCSC.
### Here is an example of how to query the server on the US west coast from
### the Unix command line:
###
###   mysql --user=genome --host=genome-mysql.soe.ucsc.edu mm10 \
###         -e "select count(*) from knownToLocusLink;"
###
### By default UCSC_dbselect() uses the server located on the US west coast.
UCSC_dbselect <- function(dbname, from, columns=NULL, where=NULL, MoreSQL=NULL,
                          host="genome-mysql.soe.ucsc.edu", port=3306,
                          try.blob.as.list=FALSE)
{
    S4Vectors:::load_package_gracefully("DBI", "by UCSC_dbselect()")
    S4Vectors:::load_package_gracefully("RMariaDB", "by UCSC_dbselect()")
    stopifnot(isTRUEorFALSE(try.blob.as.list))

    SQL <- .make_SQL_SELECT(from, columns=columns, where=where, MoreSQL=MoreSQL)
    dbconn <- DBI::dbConnect(RMariaDB::MariaDB(), dbname=dbname,
                                                  username="genome",
                                                  host=host,
                                                  port=port)
    on.exit(DBI::dbDisconnect(dbconn))
    ans <- DBI::dbGetQuery(dbconn, SQL)

    ## Undo some of the unfortunate things RMariaDB does to the result.
    ans <- .fix_RMariaDB_blobs(ans, try.blob.as.list=try.blob.as.list)
    colnames(ans) <- .unmangle_RMariaDB_colnames(colnames(ans))
    ans
}

test_that("UCSC_dbselect()", {
    result <- UCSC_dbselect("gorGor6", "ncbiRefSeqCurated")
    expect_true(is.data.frame(result))
    EXPECTED_COLNAMES <- c("bin", "name", "chrom", "strand",
                           "txStart", "txEnd", "cdsStart", "cdsEnd",
                           "exonCount", "exonStarts", "exonEnds", "score",
                           "name2", "cdsStartStat", "cdsEndStat", "exonFrames")
    expect_identical(colnames(result), EXPECTED_COLNAMES)
    expect_true(nrow(result) >= 400L)

    result <- UCSC_dbselect("gorGor6", "gap")
    expect_true(is.data.frame(result))
    EXPECTED_COLNAMES <- c("bin", "chrom", "chromStart", "chromEnd", "ix" ,                                "n", "size", "type", "bridge")
    expect_identical(colnames(result), EXPECTED_COLNAMES)
    expect_true(nrow(result) >= 800L)

    columns <- c("chrom", "chromStart", "chromEnd", "type")
    where <- "chrom='chrX' AND type='contig'"
    result <- UCSC_dbselect("gorGor6", "gap", columns=columns, where=where)
    expect_true(is.data.frame(result))
    expect_identical(colnames(result), columns)
    expect_true(nrow(result) >= 2L)
})

---
title: "The _UCSC.utils_ package"
author:
- name: Hervé Pagès
  affiliation: Fred Hutch Cancer Center, Seattle, WA
date: "Compiled `r doc_date()`; Modified 14 January 2025"
package: UCSC.utils
vignette: |
  %\VignetteIndexEntry{The UCSC.utils package}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
output:
  BiocStyle::html_document
---


# Introduction

`r Biocpkg("UCSC.utils")` is an infrastructure package that provides a
small set of low-level utilities to retrieve data from the UCSC Genome
Browser. Most functions in the package access the data via the UCSC REST
API but some of them query the UCSC MySQL server directly.

Note that the primary purpose of the package is to support higher-level
functionalities implemented in downstream packages like
`r Biocpkg("GenomeInfoDb")` or `r Biocpkg("txdbmaker")`.


# Installation

Like any other Bioconductor package, `r Biocpkg("UCSC.utils")` should always
be installed with `BiocManager::install()`:
```{r, eval=FALSE}
if (!require("BiocManager", quietly=TRUE))
    install.packages("BiocManager")
BiocManager::install("UCSC.utils")
```
However, note that `r Biocpkg("UCSC.utils")` will typically get automatically
installed as a dependency of other Bioconductor packages, so explicit
installation of the package is usually not needed.


# Functions defined in the package

## list_UCSC_genomes()

```{r list_UCSC_genomes}
suppressPackageStartupMessages(library(UCSC.utils))

list_UCSC_genomes("cat")
```

See `?list_UCSC_genomes` for more information and additional examples.

## get_UCSC_chrom_sizes()

```{r get_UCSC_chrom_sizes}
felCat9_chrom_sizes <- get_UCSC_chrom_sizes("felCat9")
head(felCat9_chrom_sizes)
```

See `?get_UCSC_chrom_sizes` for more information and additional examples.

## list_UCSC_tracks()

```{r list_UCSC_tracks}
list_UCSC_tracks("felCat9", group="varRep")
```

See `?list_UCSC_tracks` for more information and
additional examples.

## fetch_UCSC_track_data()

```{r fetch_UCSC_track_data}
mm9_cytoBandIdeo <- fetch_UCSC_track_data("mm9", "cytoBandIdeo")
head(mm9_cytoBandIdeo)
```

See `?fetch_UCSC_track_data` for more information and
additional examples.

## UCSC_dbselect()

Retrieve a full SQL table:
```{r UCSC_dbselect}
felCat9_refGene <- UCSC_dbselect("felCat9", "refGene")
head(felCat9_refGene)
```

Or retrieve a subset of it:
```{r UCSC_dbselect2}
columns <- c("chrom", "strand", "txStart", "txEnd", "exonCount", "name2")
UCSC_dbselect("felCat9", "refGene", columns=columns, where="chrom='chrA1'")
```

Note that `UCSC_dbselect` is an alternative to `fetch_UCSC_track_data`
that is more efficient and gives the user more control on what data to
retrieve exactly from the server. However, the downside of it is that
it does not work with all tracks!

See `?UCSC_dbselect` for more information and additional examples.


# Session information

```{r}
sessionInfo()
```

\name{get_UCSC_chrom_sizes}

\alias{get_UCSC_chrom_sizes}

\title{List UCSC chromosome sizes}

\description{
  Get the chromosome sizes of a given UCSC genome.
}

\usage{
get_UCSC_chrom_sizes(genome, api.url=UCSC.api.url(), recache=FALSE)
}

\arguments{
  \item{genome}{
    A single string specifying the name of a UCSC genome e.g. \code{"hs1"},
    \code{"mm39"}, or \code{"sacCer3"}. See \code{?\link{list_UCSC_genomes}}
    for how to get the list of valid UCSC genome names.
  }
  \item{api.url}{
    The URL of the UCSC API. By default, the URL returned by
    \code{UCSC.api.url()} is used. Note that what \code{UCSC.api.url()}
    returns is controlled by a global option. See \code{?\link{UCSC.api.url}}
    for more information.
  }
  \item{recache}{
    \code{get_UCSC_chrom_sizes()} uses a cache mechanism so the information
    retrieved for a given genome only gets downloaded once during the
    current R session (note that the caching is done in memory so
    cached information does NOT persist across sessions).
    Setting \code{recache} to \code{TRUE} forces a new download (and
    recaching) of the chromosome sizes for the specified genome.
  }
}

\value{
  A named numeric vector. The names on the vector are the UCSC
  chromosomes/sequences. The vector values are the corresponding lengths.

  Note that the vector is not sorted in any particular order. In particular
  there's not guarantee that the chromosomes will precede the scaffolds.
}

\seealso{
  \itemize{
    \item \code{\link{list_UCSC_genomes}} to get the list of UCSC genomes.

    \item The \code{\link[GenomeInfoDb]{Seqinfo}} constructor function
          in the \pkg{GenomeInfoDb} package for an alternate (higher level)
          way of retrieving the chromosome information of a given NCBI
          assembly or UCSC genome.

    \item \code{\link{list_UCSC_tracks}} to get the list of tracks and
          associated primary tables for a given UCSC genome.

    \item \code{\link{UCSC.api.url}} for how to use an alternative
          UCSC API URL by default.
  }
}

\examples{
get_UCSC_chrom_sizes("ce2")

get_UCSC_chrom_sizes("hg38")
}

\keyword{manip}

\name{fetch_UCSC_track_data}

\alias{fetch_UCSC_track_data}

\title{Fetch UCSC track data}

\description{
  Fetch the track data for a given UCSC genome/track.
}

\usage{
fetch_UCSC_track_data(genome, primary_table, api.url=UCSC.api.url())
}

\arguments{
  \item{genome}{
    A single string specifying the name of a UCSC genome e.g. \code{"hs1"},
    \code{"danRer11"}, or \code{"wuhCor1"}. See \code{?\link{list_UCSC_genomes}}
    for how to get the list of valid UCSC genome names.
  }
  \item{primary_table}{
    A single string specifying the name of the primary table associated
    with the track from which to fetch the data.
    See \code{?\link{list_UCSC_tracks}} for how to get the list of tracks
    and associated primary tables for a given UCSC genome.
  }
  \item{api.url}{
    The URL of the UCSC API. By default, the URL returned by
    \code{UCSC.api.url()} is used. Note that what \code{UCSC.api.url()}
    returns is controlled by a global option. See \code{?\link{UCSC.api.url}}
    for more information.
  }
}

\value{
  A data frame.
}

\seealso{
  \itemize{
    \item \code{\link{list_UCSC_genomes}} to get the list of UCSC genomes.

    \item \code{\link{list_UCSC_tracks}} to get the list of tracks and
          associated primary tables for a given UCSC genome.

    \item \code{\link{UCSC_dbselect}} for a more efficient and more flexible
          way to retrieve data directly from the UCSC MariaDB server.
  }
}

\examples{
gorGor6_gap_data <- fetch_UCSC_track_data("gorGor6", "gap")
head(gorGor6_gap_data)

## --- Comparison with UCSC_dbselect() ---

gorGor6_gap_data2 <- UCSC_dbselect("gorGor6", "gap")

## Easy sanity checks.
stopifnot(
  identical(dim(gorGor6_gap_data), dim(gorGor6_gap_data2)),
  identical(colnames(gorGor6_gap_data), colnames(gorGor6_gap_data2))
)

## But the two data frames are not identical:
identical(gorGor6_gap_data, gorGor6_gap_data2)  # FALSE!

## However, they have the same content. Ony reason they're not
## identical is because their rows are not in the same order.
## Let's reorder the rows in the two data frames by genomic location,
## that is, first by chromosome, then by chromStart, and finally
## by chromEnd:
sort_rows <- function(gap_data) {
  oo <- order(gap_data$chrom, gap_data$chromStart, gap_data$chromEnd)
  gap_data <- gap_data[oo, ]
  rownames(gap_data) <- NULL
  gap_data
}

gorGor6_gap_data <- sort_rows(gorGor6_gap_data)
gorGor6_gap_data2 <- sort_rows(gorGor6_gap_data2)

## Now the two data frames are identical:
stopifnot(identical(gorGor6_gap_data, gorGor6_gap_data2))
}

\keyword{manip}

\name{UCSC.api.url}

\alias{UCSC.api.url}

\title{Get or set the default UCSC API URL}

\description{
  Convenience helper for getting or setting global option \code{UCSC.api.url}.
}

\usage{
UCSC.api.url(api.url=NULL)
}

\arguments{
  \item{api.url}{
    A single string containing the URL to the alternative UCSC API to use
    by default.

    For convenience, \code{api.url} can also be one of the following aliases:
    \itemize{
      \item \code{"primary"}: alias for \code{"https://api.genome.ucsc.edu"}
            (primary URL, US West Coast);
      \item \code{""}: same as \code{"primary"};
      \item \code{"europe"}: alias for
            \code{"https://genome-euro.ucsc.edu/cgi-bin/hubApi"}
            (Europe mirror);
      \item \code{"asia"}: alias for
            \code{"https://genome-asia.ucsc.edu/cgi-bin/hubApi"}
            (Asia mirror).
    }
  }
}

\details{
  Various functions in the \pkg{UCSC.utils} package query the UCSC REST API.
  This is the case for example for \code{\link{list_UCSC_genomes}},
  \code{\link{get_UCSC_chrom_sizes}}, \code{\link{list_UCSC_tracks}}
  and more.

  Global option \code{UCSC.api.url} controls the UCSC API URL
  that these functions use y default. The option is set to
  \code{"https://api.genome.ucsc.edu"} (primary URL, US West Coast)
  at package startup.

  \code{UCSC.api.url()} and \code{UCSC.api.url(some_url)} are provided
  as convenient ways of doing \code{getOption("UCSC.api.url")}
  and \code{options(UCSC.api.url=some_url)}, respectively.
}

\value{
  When called with no argument, \code{UCSC.api.url()} returns
  \code{getOption("UCSC.api.url")}.

  When passed an URL, \code{UCSC.api.url(some_url)} returns the
  \emph{previous} URL, that is, the UCSC API URL that was previously
  used by default. Note that the previous URL is returned invisibly.
}

\seealso{
  \itemize{
    \item \url{https://genome.ucsc.edu/goldenPath/help/api.html#Mirrors}
          a list of alternative UCSC API URLs.

    \item \code{\link{list_UCSC_genomes}} to get the list of UCSC genomes.
  }
}

\examples{
UCSC.api.url()  # current default value of the UCSC API URL
get_UCSC_chrom_sizes("ce11", recache=TRUE)

## Temporarily use the mirror in Europe:
previous_url <- UCSC.api.url("europe")
UCSC.api.url()  # new default value of the UCSC API URL

get_UCSC_chrom_sizes("ce11", recache=TRUE)

## Restore previous default value:
UCSC.api.url(previous_url)
UCSC.api.url()
}

\keyword{utilities}

\name{list_UCSC_tracks}

\alias{list_UCSC_tracks}

\title{List UCSC tracks and associated primary tables}

\description{
  Get the list of tracks and associated primary tables for a given
  UCSC genome.
}

\usage{
list_UCSC_tracks(genome, group=NULL,
                 api.url=UCSC.api.url(), recache=FALSE)
}

\arguments{
  \item{genome}{
    A single string specifying the name of a UCSC genome e.g. \code{"hg38"},
    \code{"mm39"}, or \code{"sacCer3"}. See \code{?\link{list_UCSC_genomes}}
    for how to get the list of valid UCSC genome names.
  }
  \item{group}{
    \code{NULL} or a single string specifying the group of tracks
    to return. By default, all tracks are returned.
    Passing \code{group=NA} is accepted and will return only rows
    associated with tracks that don't belong to any group.
  }
  \item{api.url}{
    The URL of the UCSC API. By default, the URL returned by
    \code{UCSC.api.url()} is used. Note that what \code{UCSC.api.url()}
    returns is controlled by a global option. See \code{?\link{UCSC.api.url}}
    for more information.
  }
  \item{recache}{
    \code{list_UCSC_tracks()} uses a cache mechanism
    so the information retrieved for a given genome only gets downloaded
    once during the current R session (note that the caching is done in
    memory so cached information does NOT persist across sessions).
    Setting \code{recache} to \code{TRUE} forces a new download (and
    recaching) of the list of tracks for the specified genome.
  }
}

\value{
  A data frame with 1 row per track and 5 columns:
  \code{track}, \code{primary_table}, \code{type}, \code{group},
  \code{composite_track}.

  Note that columns \code{group} and \code{composite_track} can
  contain \code{NA}s.
}

\seealso{
  \itemize{
    \item \code{\link{list_UCSC_genomes}} to get the list of UCSC genomes.

    \item \code{\link{fetch_UCSC_track_data}} to fetch the track data for
          a given UCSC genome/track.

    \item \code{\link{get_UCSC_chrom_sizes}} to get the chromosome sizes
          of a given UCSC genome.

    \item \code{\link{UCSC.api.url}} for how to use an alternative
          UCSC API URL by default.
  }
}

\examples{
## List all tracks for ce2 genome:
list_UCSC_tracks("ce2")

## List tracks in the "rna" group only:
list_UCSC_tracks("ce2", group="rna")

## Note that some tracks don't belong to any group:
list_UCSC_tracks("hg38", group=NA)
}

\keyword{manip}

\name{UCSC_dbselect}

\alias{UCSC_dbselect}

\title{Conveniently retrieve data from the UCSC MariaDB server}

\description{
  A convenience utility to retrieve data from the UCSC MariaDB server
  via simple SQL \code{SELECT} statements.

  Requires the \pkg{RMariaDB} package!
}

\usage{
UCSC_dbselect(dbname, from, columns=NULL, where=NULL, MoreSQL=NULL,
              host="genome-mysql.soe.ucsc.edu", port=3306,
              try.blob.as.list=FALSE)
}

\arguments{
  \item{dbname}{
    A single string specifying the name of the database to connect to.
    This is usually the name of a valid UCSC genome e.g. \code{"hg38"},
    \code{"mm39"}, or \code{"sacCer3"}. See \code{?\link{list_UCSC_genomes}}
    for how to get the list of valid UCSC genome names.
  }
  \item{from}{
    A single string specifying the \emph{input data}, that is, the
    \code{FROM} clause of the \code{SELECT} statement. This is typically
    the name of an SQL table, or a join-clause, or a sub-query.
  }
  \item{columns}{
    \code{NULL} (the default), or a character vector specifying the columns
    to return. By default all columns are returned.
  }
  \item{where}{
    \code{NULL} (the default), or a single string specifying the \code{WHERE}
    clause of the \code{SELECT} statement.
  }
  \item{MoreSQL}{
    \code{NULL} (the default), or a character vector containing additional
    SQL clauses e.g. \code{GROUP BY}, \code{ORDER BY}, and/or \code{LIMIT}
    clauses.
  }
  \item{host}{
    A single string specifying the name of the server's host
  }
  \item{port}{
    The TCP/IP port to use to connect to the server.
  }
  \item{try.blob.as.list}{
    By default, any column of type \code{blob} returned by \pkg{RMariaDB}
    is converted to a character vector.
    If \code{try.blob.as.list} is set to \code{TRUE}, \code{UCSC_dbselect()}
    first tries to convert any blob column to a list of integers, and falls
    back to the character vector representation if that didn't work.
  }
}

\details{
  \code{UCSC_dbselect} is an alternative to \code{\link{fetch_UCSC_track_data}}
  that is more efficient and gives the user more control on what data to
  retrieve exactly from the server.

  However, the downside is that \code{UCSC_dbselect} does not work with all
  tracks! This is because, not all track data are stored in a database:
  some tracks are actually file-based (e.g. bigBed tracks). See below for
  an example.
}

\value{
  A data frame.
}

\seealso{
  \itemize{
    \item \code{\link{list_UCSC_genomes}} to get the list of UCSC genomes.

    \item \code{\link{list_UCSC_tracks}} to get the list of tracks and
          associated primary tables for a given UCSC genome.

    \item \code{\link{fetch_UCSC_track_data}} to fetch the track data for
          a given UCSC genome/track.

    \item \url{https://genome.ucsc.edu/goldenpath/help/mysql.html} for
          more information about downloading data from UCSC MariaDB
          databases.
  }
}

\examples{
### Retrieve full "ncbiGene" table ("NCBI Genes" track) for eboVir3:
df <- UCSC_dbselect("eboVir3", "ncbiGene")
df

### The exonStarts and exonEnds columns are character vectors:
sapply(df, class)
df$exonStarts
df$exonEnds

### Use 'try.blob.as.list=TRUE' to have them returned as lists of
### integers:
df2 <- UCSC_dbselect("eboVir3", "ncbiGene", try.blob.as.list=TRUE)
sapply(df2, class)
df2$exonStarts
df2$exonEnds

### Retrieve subset of "sgpGene" table ("SGP Genes" track) for hg38:
UCSC_dbselect("hg38", "sgpGene", where="chrom='chrM'")

### Retrieve subset of "gap" table for gorGor6:
columns <- c("chrom", "chromStart", "chromEnd", "type")
where <- "chrom='chrX' AND type='contig'"
UCSC_dbselect("gorGor6", "gap", columns=columns, where=where)

### With a LEFT JOIN and a LIMIT clause:
from <- paste("ncbiRefSeq LEFT JOIN ncbiRefSeqLink",
              "ON ncbiRefSeq.name=ncbiRefSeqLink.id")
res <- UCSC_dbselect("sacCer3", from, MoreSQL="LIMIT 5")
dim(res)

## WARNING: UCSC_dbselect() does not work with all tracks!
## For example it does not work with bigBed tracks:
subset(list_UCSC_tracks("hg38", group="genes"), grepl("bigBed 9", type))
\dontrun{
UCSC_dbselect("hg38", "crisprAllTargets")  # table doesn't exist!
}
}

\keyword{manip}

\name{list_UCSC_genomes}

\alias{list_UCSC_genomes}
\alias{get_organism_for_UCSC_genome}

\title{List UCSC genomes}

\description{
  Get the list of UCSC genomes.
}

\usage{
list_UCSC_genomes(organism=NA, api.url=UCSC.api.url(), recache=FALSE)

get_organism_for_UCSC_genome(genome, api.url=UCSC.api.url(), recache=FALSE)
}

\arguments{
  \item{organism}{
    By default all UCSC genomes are returned (in a data frame).
    When \code{organism} is specified, \code{list_UCSC_genomes()}
    will only return the rows associated with the specified organism.
    \code{organism} must be supplied as a single string that will be
    used to perform a search (with \code{grep()}) on the \code{organism}
    and \code{common_name} columns of the data frame to return.
    The search is case-insensitive.
  }
  \item{api.url}{
    The URL of the UCSC API. By default, the URL returned by
    \code{UCSC.api.url()} is used. Note that what \code{UCSC.api.url()}
    returns is controlled by a global option. See \code{?\link{UCSC.api.url}}
    for more information.
  }
  \item{recache}{
    \code{list_UCSC_genomes()} uses a cache mechanism so the list of genomes
    only gets downloaded once during the current R session (note that the
    caching is done in memory so cached information does NOT persist across
    sessions).
    Setting \code{recache} to \code{TRUE} forces a new download (and
    recaching) of the list of genomes.

    In the case of \code{get_organism_for_UCSC_genome()}, the supplied
    \code{recache} value is just passed down to the internal call
    to \code{list_UCSC_genomes()}.
  }
  \item{genome}{
    A character vector of valid UCSC genomes e.g. \code{"hg38"},
    \code{"mm39"}, or \code{"sacCer3"}.
  }
}

\value{
  For \code{list_UCSC_genomes}: A data frame with 1 row per genome
  and 5 columns: \code{organism}, \code{genome}, \code{common_name},
  \code{tax_id}, \code{description}.

  For \code{get_organism_for_UCSC_genome}: A named character of the same
  length as the input containing the scientific names of the organisms
  associated with the supplied UCSC genomes.
}

\seealso{
  \itemize{
    \item \url{https://genome.ucsc.edu/FAQ/FAQreleases.html} for the online
          HTML page that lists all UCSC genome releases, including archived
          ones.

    \item \code{\link[GenomeInfoDb]{registered_UCSC_genomes}} in the
          \pkg{GenomeInfoDb} package for a similar function that returns
          only UCSC genomes registered in the \pkg{GenomeInfoDb} package.

    \item \code{\link{get_UCSC_chrom_sizes}} to get the chromosome sizes
          of a given UCSC genome.

    \item \code{\link{list_UCSC_tracks}} to get the list of tracks and
          associated primary tables for a given UCSC genome.

    \item \code{\link{UCSC.api.url}} for how to use an alternative
          UCSC API URL by default.
  }
}

\examples{
list_UCSC_genomes("human")

list_UCSC_genomes("pacos")

get_organism_for_UCSC_genome(c("ce11", "xenTro10", "mpxvRivers"))
}

\keyword{manip}
